from sacad.sources.amazoncd import AmazonCdCoverSource, AmazonCdCoverSourceResult
from sacad.sources.amazondigital import AmazonDigitalCoverSource, AmazonDigitalCoverSourceResult
from sacad.sources.lastfm import LastFmCoverSource, LastFmCoverSourceResult
from sacad.sources.google_images import GoogleImagesWebScrapeCoverSource, GoogleImagesCoverSourceResult
